/*
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */


#include <stdint.h>
#include <sys/types.h>


#include "libavutil/common.h"
#include "libavutil/opt.h"
#include "libavutil/internal.h"

#include "avcodec.h"
#include "internal.h"
#include "profiles.h"
#include "get_bits.h"

static av_cold int prores_raw_init(AVCodecContext *avctx)
{
    avctx->bits_per_raw_sample = 16;

    switch (avctx->codec_tag) {
    case MKTAG('a','p','r','n'):
        avctx->profile = FF_PROFILE_PRORES_RAW_STANDARD;
        break;
    case MKTAG('a','p','r','h'):
        avctx->profile = FF_PROFILE_PRORES_RAW_HQ;
        break;
    default:
        avctx->profile = FF_PROFILE_UNKNOWN;
        av_log(avctx, AV_LOG_WARNING, "Unknown prores raw profile %d\n", avctx->codec_tag);
    }
    
    avctx->pix_fmt = AV_PIX_FMT_RGBA64;
    
    return 0;
}

static int prores_raw_frame_header(const uint8_t *buf, const int data_size, AVCodecContext *avctx)
{
    int hdr_size, width, height, flags;
    int crop_left, crop_right, crop_top, crop_bottom;
    int version;
    const uint8_t *ptr;

    hdr_size = AV_RB16(buf);
    if (hdr_size > data_size) {
        av_log(avctx, AV_LOG_ERROR, "error, wrong header size\n");
        return AVERROR_INVALIDDATA;
    }

    version = AV_RB16(buf + 2);

    width  = AV_RB16(buf + 8);
    height = AV_RB16(buf + 10);
    crop_left = buf[12];
    crop_right = buf[13];
    crop_top = buf[14];
    crop_bottom = buf[15];
    
    width -= crop_left + crop_right;
    height -= crop_top + crop_bottom;

    if (width != avctx->width || height != avctx->height) {
        int ret;

        av_log(avctx, AV_LOG_WARNING, "picture resolution change: %dx%d -> %dx%d\n",
               avctx->width, avctx->height, width, height);
        if ((ret = ff_set_dimensions(avctx, width, height)) < 0)
            return ret;
    }

    return hdr_size;
}

static int prores_raw_frame(AVCodecContext *avctx, void *data, int *got_frame, AVPacket *avpkt)
{
    AVFrame *frame = data;
    const uint8_t *buf = avpkt->data;
    int buf_size = avpkt->size;
    int frame_hdr_size, pic_size, ret;

    if (buf_size < 28 || AV_RL32(buf + 4) != AV_RL32("prrf")) {
        av_log(avctx, AV_LOG_ERROR, "invalid frame header\n");
        return AVERROR_INVALIDDATA;
    }

    buf += 8;
    buf_size -= 8;

    frame_hdr_size = prores_raw_frame_header(buf, buf_size, avctx);
    if (frame_hdr_size < 0)
        return frame_hdr_size;
    
    if (avctx->flags & AV_CODEC_DECODE_HEADER_ONLY) {
        av_log(avctx, AV_LOG_DEBUG, "decode_frame stop after header decoded\n");
        return avpkt->size;
    }
    
    return -1;
}

static av_cold int prores_raw_close(AVCodecContext *avctx)
{
    return 0;
}

AVCodec ff_prores_raw_decoder = {
    .name           = "prores_raw",
    .long_name      = NULL_IF_CONFIG_SMALL("ProRes RAW (iCodec Pro)"),
    .type           = AVMEDIA_TYPE_VIDEO,
    .id             = AV_CODEC_ID_PRORES_RAW,
    .init           = prores_raw_init,
    .close          = prores_raw_close,
    .decode         = prores_raw_frame,
    .capabilities   = AV_CODEC_CAP_DR1,
    .profiles       = NULL_IF_CONFIG_SMALL(ff_prores_raw_profiles),
};
